import logging
import os
import socket
from io import BytesIO
from typing import TypedDict

import cbor2

from cbor2 import dumps, CBORDecoder

from rewrite_remote.handlers.project_helper import (
    find_python_files,
    parse_python_sources,
)

from rewrite_remote.remoting import (
    OK,
    RemotingMessageType,
)

from rewrite_remote.remote_utils import COMMAND_END
from rewrite_remote.remoting import RemotingContext
from rewrite_remote.handlers.handler_helpers import respond_with_error
from rewrite_remote.remoting import RemotingMessenger


class ParseProjectSourcesArgs(TypedDict):
    project_file_path: str  # The path to the individual pyproject.toml
    root_project_file_path: str  #  The path to the root pyproject.toml
    repository_dir: str  # The path to the root repository directory


def decode_parse_project_sources_args(
    decoder: CBORDecoder,
) -> ParseProjectSourcesArgs:
    """
    Decodes the arguments (order matters and must match the order encoded)
    """
    project_file_path = str(decoder.decode())
    root_project_file_path = str(decoder.decode())
    repository_dir = str(decoder.decode())

    return {
        "project_file_path": project_file_path,
        "root_project_file_path": root_project_file_path,
        "repository_dir": repository_dir,
    }


def parse_project_sources_handler(
    stream: BytesIO, sock: socket.socket, remoting_ctx: RemotingContext
) -> None:
    remoting_ctx.reset()

    # Read input from stream
    try:
        data = stream.read()
        decoder = CBORDecoder(BytesIO(data))
        args = decode_parse_project_sources_args(decoder)
        project_file_path = args.get("project_file_path")
        root_project_file_path = args.get("root_project_file_path")
        repository_dir = args.get("repository_dir")
    except Exception as e:  # pylint: disable=broad-except
        respond_with_error(f"Failed to decode arguments: {e}", sock)
        return

    if project_file_path is None:
        respond_with_error("recipe_name is required", sock)
        return

    # Log the request
    logging.info(
        "[Server] Handling parse-project-sources request: {"
        "project_file_path: %s, root_project_file_path: %s, repository_dir: %s}",
        project_file_path,
        root_project_file_path,
        repository_dir,
    )

    # Find all python files in the project
    base_dir = os.path.dirname(project_file_path)
    python_files = find_python_files(base_dir)
    source_files = parse_python_sources(python_files)

    # Write the response
    response_stream = BytesIO()
    cbor2.dump(RemotingMessageType.Response, response_stream)
    cbor2.dump(OK, response_stream)
    for source_file in source_files:
        logging.info("Sending ${source_file.source_path}")
        RemotingMessenger.send_tree(remoting_ctx, response_stream, source_file, None)
    cbor2.dump(COMMAND_END, response_stream)
    sock.sendall(response_stream.getvalue())

    logging.info("[Server] Request completed.")



# type: ignore
import socket
import struct
from enum import Enum, auto
from io import BytesIO
from typing import (
    Callable,
    List,
    Optional,
    TypeVar,
    Dict,
    TYPE_CHECKING,
    Iterable,
    cast,
    Type,
    Any,
)

from cbor2 import CBORDecoder, CBORDecodeValueError

from .event import EventType

if TYPE_CHECKING:
    from rewrite_remote.receiver import ReceiverContext, DetailsReceiver

T = TypeVar("T")
I = TypeVar("I")


class Operation(Enum):
    Add = auto()
    Delete = auto()
    NoChange = auto()
    Update = auto()
    Move = auto()


def receive_nodes(
    before: Optional[List[T]],
    details: Callable[[Optional[T], Optional[str], "ReceiverContext"], T],
    ctx: "ReceiverContext",
) -> Optional[List[T]]:
    list_event = ctx.receiver.receive_value(list)
    if list_event.event_type == EventType.NoChange:
        return before
    elif list_event.event_type == EventType.Delete:
        return None
    elif list_event.event_type == EventType.Add:
        after_size = list_event.msg
        if after_size is None:
            after_size = 0
        after = [None] * after_size  # Preallocate list
        for i in range(after_size):
            diff_event = ctx.receiver.receive_node()
            if diff_event.event_type == EventType.Add:
                after[i] = details(None, diff_event.concrete_type, ctx)  # type: ignore
            elif diff_event.event_type == EventType.NoChange:
                after[i] = None  # Or some default value
            else:
                raise NotImplementedError(f"Unexpected operation: {diff_event.event_type}")
        return after  # type: ignore
    elif list_event.event_type == EventType.Update:
        return _receive_updated_nodes(before, list_event.msg, details, ctx)  # type: ignore
    else:
        raise NotImplementedError(f"Unexpected operation: {list_event.event_type}")


def _receive_updated_nodes(
    before: List[T],
    after_size: int,
    details: "DetailsReceiver[T]",
    ctx: "ReceiverContext",
) -> List[T]:
    modified = False
    after_list = before
    evt = ctx.receiver.receive_node()
    if evt.event_type != EventType.StartList:
        raise ValueError(f"Expected start list event: {evt.event_type}")

    before_idx = 0
    while True:
        evt = ctx.receiver.receive_node()
        if evt.event_type in (EventType.NoChange, EventType.EndList):
            break

        if evt.event_type in (
            EventType.Delete,
            EventType.Update,
            EventType.Add,
        ):
            if not modified:
                after_list = _copy_range(before, before_idx)
                modified = True

        if evt.event_type == EventType.NoChange:
            if modified:
                after_list.append(before[before_idx])
            before_idx += 1
        elif evt.event_type == EventType.Delete:
            before_idx += 1
        elif evt.event_type == EventType.Update:
            after_list.append(details.receive_details(before[before_idx], evt.concrete_type, ctx))
            before_idx += 1
        elif evt.event_type == EventType.Add:
            after_list.append(details.receive_details(None, evt.concrete_type, ctx))

        if evt.event_type == EventType.EndList:
            break

    return after_list[:after_size] if len(after_list) > after_size else after_list


def receive_values(
    before: Optional[List[T]], type: Type[Any], ctx: "ReceiverContext"
) -> Optional[List[T]]:
    list_event = ctx.receiver.receive_value(list)
    if list_event.event_type == EventType.NoChange:
        return before
    elif list_event.event_type == EventType.Delete:
        return None
    elif list_event.event_type == EventType.Add:
        after_size = list_event.msg
        if after_size is None:
            after_size = 0
        after = [None] * after_size  # Preallocate list
        for i in range(after_size):
            diff_event = ctx.receiver.receive_value(type)
            if diff_event.event_type == EventType.Add:
                after[i] = diff_event.msg
            elif diff_event.event_type == EventType.NoChange:
                after[i] = None  # Or some default value
            else:
                raise NotImplementedError(f"Unexpected operation: {diff_event.event_type}")
        return after  # type: ignore
    elif list_event.event_type == EventType.Update:
        return _receive_updated_values(before, list_event.msg, type, ctx)  # type: ignore
    else:
        raise NotImplementedError(f"Unexpected operation: {list_event.event_type}")


def _receive_updated_values(
    before: List[T], after_size: int, type: Type[Any], ctx: "ReceiverContext"
) -> List[T]:
    modified = False
    after_list = before
    evt = ctx.receiver.receive_node()
    if evt.event_type != EventType.StartList:
        raise ValueError(f"Expected start list event: {evt.event_type}")

    before_idx = 0
    while True:
        evt = ctx.receiver.receive_value(type)
        if evt.event_type in (EventType.NoChange, EventType.EndList):
            break

        if evt.event_type in (
            EventType.Delete,
            EventType.Update,
            EventType.Add,
        ):
            if not modified:
                after_list = _copy_range(before, before_idx)
                modified = True

        if evt.event_type == EventType.NoChange:
            if modified:
                after_list.append(before[before_idx])
            before_idx += 1
        elif evt.event_type == EventType.Delete:
            before_idx += 1
        elif evt.event_type in (EventType.Update, EventType.Add):
            after_list.append(cast(T, evt.msg))
            if evt.event_type == EventType.Update:
                before_idx += 1

        if evt.event_type == EventType.EndList:
            break

    return after_list[:after_size] if len(after_list) > after_size else after_list


def _copy_range(before: Iterable[T], j: int) -> List[T]:
    if isinstance(before, list):
        return before[:j]
    elif hasattr(
        before, "getrange"
    ):  # If the object has a 'getrange' method (e.g., an immutable list)
        return cast(List[T], before.getrange(0, j))
    else:
        return list(before)[:j]


def calculate_list_diff(
    before: List[T],
    after: List[T],
    id_function: Callable[[T], I],
    consumer: Callable[[Operation, int, int, Optional[T], Optional[T]], None],
) -> None:
    before_idx, after_idx = 0, 0
    before_size, after_size = len(before), len(after)
    after_map = None

    while before_idx < before_size or after_idx < after_size:
        # Check if we've reached the end of either of the lists
        if before_idx >= before_size:
            consumer(Operation.Add, -1, after_idx, None, after[after_idx])
            after_idx += 1
            continue
        elif after_idx >= after_size:
            consumer(Operation.Delete, before_idx, -1, before[before_idx], None)
            before_idx += 1
            continue

        if before[before_idx] == after[after_idx]:
            consumer(
                Operation.NoChange,
                before_idx,
                after_idx,
                before[before_idx],
                after[after_idx],
            )
            before_idx += 1
            after_idx += 1
        else:
            before_id = id_function(before[before_idx])
            after_id = id_function(after[after_idx])

            if before_id == after_id:
                consumer(
                    Operation.Update,
                    before_idx,
                    after_idx,
                    before[before_idx],
                    after[after_idx],
                )
                before_idx += 1
                after_idx += 1
            else:
                if after_map is None:
                    after_map = create_index_map(after, after_idx, id_function)

                # If elements at current indices are not equal, figure out the operation
                if before_id not in after_map:
                    consumer(
                        Operation.Delete,
                        before_idx,
                        -1,
                        before[before_idx],
                        None,
                    )
                    before_idx += 1
                else:
                    consumer(Operation.Add, -1, after_idx, None, after[after_idx])
                    after_idx += 1


def create_index_map(lst: List[T], from_index: int, id_function: Callable[[T], I]) -> Dict[I, int]:
    result = {}
    for i in range(from_index, len(lst)):
        result[id_function(lst[i])] = i
    return result


def _decode_length(
    decoder: CBORDecoder, subtype: int, allow_indefinite: bool = False
) -> Optional[int]:
    if subtype < 24:
        return subtype
    elif subtype == 24:
        return decoder.read(1)[0]
    elif subtype == 25:
        return cast(int, struct.unpack(">H", decoder.read(2))[0])
    elif subtype == 26:
        return cast(int, struct.unpack(">L", decoder.read(4))[0])
    elif subtype == 27:
        return cast(int, struct.unpack(">Q", decoder.read(8))[0])
    elif subtype == 31 and allow_indefinite:
        return None
    else:
        raise CBORDecodeValueError(f"unknown unsigned integer subtype 0x{subtype:x}")


def decode_array_start(decoder: CBORDecoder) -> Optional[int]:
    initial_byte = decoder.read(1)[0]
    major_type = initial_byte >> 5
    assert major_type == 4
    subtype = initial_byte & 31
    return _decode_length(decoder, subtype, allow_indefinite=True)


COPY_BUFFER_SIZE = 4096
COPY_BUFFER = bytearray(COPY_BUFFER_SIZE)
COMMAND_END = bytes([0x81, 0x17])


def read_to_command_end(sock: socket.socket) -> BytesIO:
    memory_stream = BytesIO()

    try:
        while True:
            bytes_read = sock.recv_into(COPY_BUFFER, len(COPY_BUFFER))
            if bytes_read == 0:
                break
            memory_stream.write(COPY_BUFFER[:bytes_read])
            if (
                bytes_read > 1
                and COPY_BUFFER[bytes_read - 2] == COMMAND_END[0]
                and COPY_BUFFER[bytes_read - 1] == COMMAND_END[1]
            ):
                break
            elif bytes_read == 1:
                original_position = memory_stream.tell()
                memory_stream.seek(-2, 1)  # Move back by 2 bytes
                if (
                    memory_stream.read(1)[0] == COMMAND_END[0]
                    and memory_stream.read(1)[0] == COMMAND_END[1]
                ):
                    memory_stream.seek(original_position)
                    break
    except socket.error as e:
        print(f"Socket error: {e}")

    memory_stream.seek(0)
    return memory_stream



# type: ignore
# Ignoring type checking for this file because there is too many errors for now

import struct
from collections import OrderedDict
from enum import Enum
from pathlib import Path
from typing import (
    Protocol,
    TypeVar,
    Optional,
    Type,
    Dict,
    Callable,
    List,
    cast,
    Iterable,
    Any,
    TYPE_CHECKING,
    get_args,
    get_origin,
)
from uuid import UUID

from _cbor2 import break_marker
from cbor2 import CBORDecoder, CBORDecodeValueError
from cbor2._decoder import major_decoders

from rewrite import (
    Markers,
    Marker,
    ParseError,
    ParseErrorVisitor,
    SearchResult,
    UnknownJavaMarker,
)
from rewrite import Tree, TreeVisitor, Cursor, FileAttributes
from . import remote_utils, type_utils
from .event import DiffEvent, EventType

if TYPE_CHECKING:
    from .remoting import RemotingContext

A = TypeVar("A")
T = TypeVar("T", bound=Tree)
V = TypeVar("V")
I = TypeVar("I")
P = TypeVar("P")


class Receiver(Protocol):
    def fork(self, context: "ReceiverContext") -> "ReceiverContext": ...

    def receive(self, before: Optional[T], ctx: "ReceiverContext") -> object: ...


class OmniReceiver(Receiver):
    def fork(self, ctx: "ReceiverContext") -> "ReceiverContext":
        raise NotImplementedError("Cannot fork OmniReceiver")

    def receive(self, before: Optional["Tree"], ctx: "ReceiverContext") -> "Tree":
        visitor = self.Visitor()
        return visitor.visit(before, ctx)

    class Visitor(TreeVisitor[Tree, "ReceiverContext"]):  # type: ignore
        def visit(
            self,
            tree: Optional[Tree],
            ctx: "ReceiverContext",
            parent: Optional[Cursor] = None,
        ) -> Optional[Tree]:
            self.cursor = Cursor(self.cursor, tree)  # type: ignore
            tree = ctx.polymorphic_receive_tree(tree)
            self.cursor = self.cursor.parent
            return tree


class TreeReceiver(Protocol):
    def receive_node(self) -> DiffEvent: ...

    def receive_value(self, expected_type: Type[Any]) -> DiffEvent: ...


class ReceiverFactory(Protocol):
    def create(self, type_name: Optional[str], ctx: "ReceiverContext") -> Tree: ...


class DetailsReceiver(Protocol[T]):
    def receive_details(
        self,
        before: Optional[T],
        type: Optional[Type[T]],
        ctx: "ReceiverContext",
    ) -> T:
        pass


class ReceiverContext:
    Registry: Dict[Type[Any], Callable[[], Receiver]] = OrderedDict()

    def __init__(
        self,
        receiver: TreeReceiver,
        visitor: Optional[TreeVisitor] = None,
        factory: Optional[ReceiverFactory] = None,
    ):
        self.receiver = receiver
        self.visitor = visitor
        self.factory = factory

    def fork(self, visitor: TreeVisitor, factory: ReceiverFactory) -> "ReceiverContext":
        return ReceiverContext(self.receiver, visitor, factory)

    def receive_any_tree(self, before: Optional[T]) -> Optional[T]:
        return cast(Optional[T], OmniReceiver().receive(before, self))

    def receive_tree(
        self,
        before: Optional[Tree],
        tree_type: Optional[str],
        ctx: "ReceiverContext",
    ) -> Tree:
        if before:
            return before.accept(self.visitor, ctx)
        else:
            if self.factory is not None:
                return self.factory.create(tree_type, ctx)
            raise ValueError("Factory is not defined")

    def polymorphic_receive_tree(self, before: Optional[Tree]) -> Optional[Tree]:
        diff_event = self.receiver.receive_node()
        if diff_event.event_type in (EventType.Add, EventType.Update):
            tree_receiver = self.new_receiver(diff_event.concrete_type or type(before).__name__)
            forked = tree_receiver.fork(self)
            return forked.receive_tree(
                None if diff_event.event_type == EventType.Add else before,
                diff_event.concrete_type,
                forked,
            )
        elif diff_event.event_type == EventType.Delete:
            return None
        else:
            return before

    def new_receiver(self, type_name: str) -> Receiver:
        type_ = type_utils.get_type(type_name)
        for entry_type, factory in self.Registry.items():
            if issubclass(type_, entry_type):
                return factory()
        raise ValueError(f"Unsupported receiver type: {type_name}")

    def receive_node(
        self,
        before: Optional[A],
        details: Callable[[Optional[A], Optional[str], "ReceiverContext"], A],
    ) -> Optional[A]:
        evt = self.receiver.receive_node()
        if evt.event_type == EventType.Delete:
            return None
        elif evt.event_type == EventType.Add:
            return details(None, evt.concrete_type, self)
        elif evt.event_type == EventType.Update:
            return details(before, evt.concrete_type, self)
        return before

    def receive_markers(
        self,
        before: Optional[Markers],
        type: Optional[str],
        ctx: "ReceiverContext",
    ) -> Markers:
        id_ = self.receive_value(getattr(before, "id", None), UUID)
        after_markers: Optional[List[Marker]] = self.receive_values(
            getattr(before, "markers", None), Marker
        )
        if before:
            return before.with_id(id_).with_markers(after_markers)
        else:
            return Markers(id_, after_markers)

    def receive_nodes(
        self,
        before: Optional[List[A]],
        details: Callable[[Optional[A], Optional[str], "ReceiverContext"], A],
    ) -> Optional[List[A]]:
        return remote_utils.receive_nodes(before, details, self)

    def receive_values(self, before: Optional[List[V]], type: Type[Any]) -> Optional[List[V]]:
        return remote_utils.receive_values(before, type, self)

    def receive_value(self, before: Optional[V], type: Type[Any]) -> Optional[V]:
        return self.receive_value0(before, type)

    def receive_value0(self, before: Optional[V], type: Type[Any]) -> Optional[V]:
        evt = self.receiver.receive_value(type)
        if evt.event_type in (EventType.Update, EventType.Add):
            return evt.msg
        elif evt.event_type == EventType.Delete:
            return None
        return before

    @staticmethod
    def register(type_: Type[Any], receiver_factory: Callable[[], Receiver]) -> None:
        ReceiverContext.Registry[type_] = receiver_factory


ValueDeserializer = Callable[[str, CBORDecoder, "DeserializationContext"], Optional[Any]]


class DefaultValueDeserializer(ValueDeserializer):
    def __call__(
        self,
        expected_type: Optional[Type[Any]],
        reader: CBORDecoder,
        context: "DeserializationContext",
    ) -> Any:
        cbor_map = reader.decode_map(subtype=None)
        error_message = "No deserializer found for: " + ", ".join(
            f"{k}: {v}" for k, v in cbor_map.items()
        )
        raise NotImplementedError(error_message)


class DeserializationContext:
    DefaultDeserializer = DefaultValueDeserializer()

    value_deserializers: Dict[str, ValueDeserializer]

    def __init__(
        self,
        remoting_context: "RemotingContext",
        value_deserializers: Optional[Dict[Type, ValueDeserializer]] = None,
    ):
        self.remoting_context = remoting_context
        self.value_deserializers = value_deserializers or {}

    def deserialize(self, expected_type: Type[Any], decoder: CBORDecoder) -> Any:
        if expected_type == UUID:
            return UUID(bytes=decoder.decode())  # type: ignore

        if expected_type == str:
            return decoder.decode()

        if expected_type == bool:
            return decoder.decode()

        if expected_type == list:
            return decoder.decode()

        if expected_type == int:
            return decoder.decode()

        if expected_type == Path:
            return Path(decoder.decode())  # type: ignore

        if expected_type == float:
            return decoder.decode()

        if isinstance(expected_type, type) and issubclass(expected_type, Enum):
            return expected_type(decoder.decode())

        initial_byte = decoder.read(1)[0]
        major_type = initial_byte >> 5
        subtype = initial_byte & 31
        concrete_type: Optional[str] = None

        # Object ID for Marker, JavaType, etc.
        if major_type == 0:
            obj_id = decoder.decode_uint(subtype)
            return self.remoting_context.get_object(obj_id)

        elif major_type == 1:
            return decoder.decode_negint(subtype)

        # arrays
        elif major_type == 4:
            if get_origin(expected_type) in (List, list):
                expected_elem_type = get_args(expected_type)[0]
                array = []
                length = _decode_length(decoder, subtype, allow_indefinite=True)
                if length:
                    for _ in range(length):
                        elem = self.deserialize(expected_elem_type, decoder)
                        array.append(elem)
                else:
                    while (
                        not (value := self.deserialize(expected_elem_type, decoder)) == break_marker
                    ):
                        array.append(value)
                return array
            else:
                concrete_type = decoder.decode()

        # objects
        elif major_type == 5:
            field_name = decoder.decode()
            assert field_name == "@c"
            concrete_type = decoder.decode()

        if concrete_type:
            if concrete_type == "org.openrewrite.marker.SearchResult":
                field_name = decoder.decode()
                if field_name == "description":
                    desc = decoder.decode()
                elif field_name == "id":
                    id_ = UUID(bytes=decoder.decode())
                field_name = decoder.decode()
                if field_name == "description":
                    desc = decoder.decode()
                elif field_name == "id":
                    id_ = UUID(bytes=decoder.decode())
                return SearchResult(id_, desc)

            if deser := self.value_deserializers.get(concrete_type):
                return deser(concrete_type, decoder, self)

            for type_, value_deserializer in self.value_deserializers.items():
                if type_ == concrete_type:
                    return value_deserializer(concrete_type, decoder, self)

            initial_byte = decoder.read(1)[0]
            major_type = initial_byte >> 5
            subtype = initial_byte & 31
            while initial_byte != 0xFF:
                field_name = decoder.decode_string(subtype)
                field_value = decoder.decode()
                initial_byte = decoder.read(1)[0]
                major_type = initial_byte >> 5
                subtype = initial_byte & 31
            pass

        elif major_type == 3:
            return decoder.decode_string(subtype)
        elif major_type == 2:
            return decoder.decode_bytestring(subtype)
        elif major_type == 7:
            return decoder.decode_special(subtype)
        else:
            return major_decoders[major_type](decoder, subtype)

        state = decoder.peek_state()

        if state in {
            cbor2.CborReaderState.BOOLEAN,
            cbor2.CborReaderState.UNSIGNED_INT,
            cbor2.CborReaderState.NEGATIVE_INT,
        }:
            result = decoder.read_int()
            obj = self.remoting_context.get_object(result)
            return obj if obj is not None else result

        if state in {
            cbor2.CborReaderState.HALF_FLOAT,
            cbor2.CborReaderState.FLOAT,
            cbor2.CborReaderState.DOUBLE,
        }:
            return decoder.read_double()

        if state == cbor2.CborReaderState.TEXT_STRING:
            str_value = decoder.read_text_string()
            if decoder.peek_state() == cbor2.CborReaderState.END_ARRAY:
                return str_value

            concrete_type = (
                str_value
                if decoder.peek_state() != cbor2.CborReaderState.END_ARRAY
                else expected_type.__name__
            )

            if concrete_type == "org.openrewrite.FileAttributes":
                map = decoder.read_cbor_map()
                return FileAttributes(None, None, None, False, False, False, 0)

            if concrete_type == "java.lang.String":
                return decoder.decode()
            if concrete_type == "java.lang.Boolean":
                return decoder.decode()
            if concrete_type == "java.lang.Integer":
                return decoder.decode()
            if concrete_type == "java.lang.Character":
                return decoder.decode()[0]  # type: ignore
            if concrete_type == "java.lang.Long":
                return decoder.decode()
            if concrete_type == "java.lang.Double":
                return decoder.decode()
            if concrete_type == "java.lang.Float":
                return decoder.decode()
            if concrete_type == "java.math.BigInteger":
                return decoder.decode()
            if concrete_type == "java.math.BigDecimal":
                return decoder.decode()

            raise NotImplementedError(f"No deserialization implemented for: {concrete_type}")

        if state == cbor2.CborReaderState.ARRAY:
            decoder.read_array_start()
            concrete_type = decoder.read_text_string()
            actual_type = type_utils.get_type(concrete_type)
            for type_, deserializer in self.value_deserializers.items():
                if issubclass(actual_type, type_):
                    return deserializer.deserialize(actual_type, decoder, self)

        if state == cbor2.CborReaderState.MAP:
            if issubclass(expected_type, Marker):
                if decoder.peek_state() == cbor2.CborReaderState.UNSIGNED_INT:
                    obj_id = decoder.read_int()
                    return self.remoting_context.get_object(obj_id)

                marker_map = ValueDeserializer.read_cbor_map(decoder)

                if "@c" not in marker_map:
                    raise NotImplementedError("Expected @c key")

                concrete_type = marker_map["@c"]

                if concrete_type in {
                    "org.openrewrite.marker.SearchResult",
                    "Rewrite.Core.Marker.SearchResult",
                }:
                    desc = marker_map.get("description", None)
                    marker = SearchResult(UUID(bytes=marker_map["id"]), desc)
                else:
                    marker = UnknownJavaMarker(UUID(bytes=marker_map["id"]), marker_map)

                if "@ref" in marker_map:
                    self.remoting_context.add(marker_map["@ref"], marker)

                return marker

            decoder.read_map_start()
            if decoder.read_text_string() != "@c":
                raise NotImplementedError("Expected @c key")
            concrete_type = decoder.read_text_string()
            actual_type = type_utils.get_type(concrete_type)
            for type_, deserializer in self.value_deserializers.items():
                if issubclass(actual_type, type_):
                    return deserializer.deserialize(actual_type, decoder, self)
        raise NotImplementedError(f"No deserialization implemented for: {expected_type}")


class JsonReceiver(TreeReceiver):
    DEBUG = False

    def __init__(self, stream, context: DeserializationContext):
        super().__init__()
        self._stream = stream
        self._decoder = CBORDecoder(self._stream)
        self._context = context
        self._count = 0

    def receive_node(self):
        array = self._decoder.decode()
        if isinstance(array, list):
            event_type = EventType(cast(int, array[0]))
            msg = None
            concrete_type = None

            if event_type in {EventType.Add, EventType.Update}:
                if event_type == EventType.Add and len(array) > 1 and isinstance(array[1], str):
                    concrete_type = array[1]

            elif event_type not in {
                EventType.Delete,
                EventType.NoChange,
                EventType.StartList,
                EventType.EndList,
            }:
                raise NotImplementedError(event_type)

            if self.DEBUG:
                print(f"[{self._count}] {DiffEvent(event_type, concrete_type, msg)}")
            self._count += 1
            return DiffEvent(event_type, concrete_type, msg)

        else:
            raise NotImplementedError(f"Unexpected state: {type(array)}")

    def receive_value(self, expected_type: Type):
        length = remote_utils.decode_array_start(self._decoder)
        event_type = EventType(self._decoder.decode())
        msg = None
        concrete_type = None

        if event_type in {EventType.Add, EventType.Update}:
            if bool(expected_type) and issubclass(expected_type, (list, Iterable)):
                # special case for list events
                msg = self._decoder.decode()
            else:
                msg = self._context.deserialize(expected_type, self._decoder)

        elif event_type not in {
            EventType.Delete,
            EventType.NoChange,
            EventType.StartList,
            EventType.EndList,
        }:
            raise NotImplementedError(event_type)

        if length is None:
            array_end = self._decoder.decode()
            # assert array_end == break_marker

        if self.DEBUG:
            print(f"[{self._count}] {DiffEvent(event_type, concrete_type, msg)}")
        self._count += 1
        return DiffEvent(event_type, concrete_type, msg)


def _decode_length(
    decoder: CBORDecoder, subtype: int, allow_indefinite: bool = False
) -> Optional[int]:
    if subtype < 24:
        return subtype
    elif subtype == 24:
        return decoder.read(1)[0]
    elif subtype == 25:
        return cast(int, struct.unpack(">H", decoder.read(2))[0])
    elif subtype == 26:
        return cast(int, struct.unpack(">L", decoder.read(4))[0])
    elif subtype == 27:
        return cast(int, struct.unpack(">Q", decoder.read(8))[0])
    elif subtype == 31 and allow_indefinite:
        return None
    else:
        raise CBORDecodeValueError(f"unknown unsigned integer subtype 0x{subtype:x}")

class ParseErrorReceiver(Receiver):
    def fork(self, ctx):
        return ctx.fork(self.Visitor(), self.Factory())

    def receive(self, before, ctx):
        forked = self.fork(ctx)
        return forked.visitor.visit(before, forked)

    class Visitor(ParseErrorVisitor):
        def visit(self, tree, ctx, parent: Optional[Cursor] = None):
            self.cursor = Cursor(self.cursor, tree)
            tree = ctx.receive_node(tree, ctx.receive_tree)
            self.cursor = self.cursor.parent
            return tree

        def visit_parse_error(self, parse_error, ctx):
            parse_error = parse_error.with_id(ctx.receive_value(parse_error.id))
            parse_error = parse_error.with_markers(
                ctx.receive_node(parse_error.markers, ctx.receive_markers)
            )
            parse_error = parse_error.with_source_path(ctx.receive_value(parse_error.source_path))
            parse_error = parse_error.with_file_attributes(
                ctx.receive_value(parse_error.file_attributes)
            )
            parse_error = parse_error.with_charset_name(ctx.receive_value(parse_error.charset_name))
            parse_error = parse_error.with_charset_bom_marked(
                ctx.receive_value(parse_error.charset_bom_marked)
            )
            parse_error = parse_error.with_checksum(ctx.receive_value(parse_error.checksum))
            parse_error = parse_error.with_text(ctx.receive_value(parse_error.text))
            # parse_error = parse_error.with_erroneous(ctx.receive_tree(parse_error.erroneous))
            return parse_error

    class Factory(ReceiverFactory):
        def create(self, type_, ctx):
            if type_ in [
                "rewrite.parser.ParseError",
                "org.openrewrite.tree.ParseError",
            ]:
                return ParseError(
                    ctx.receive_value(None),
                    ctx.receive_node(None, ctx.receive_markers),
                    ctx.receive_value(None),
                    ctx.receive_value(None),
                    ctx.receive_value(None),
                    ctx.receive_value(None),
                    ctx.receive_value(None),
                    ctx.receive_value(None),
                    None,  # ctx.receive_tree(None)
                )
            raise NotImplementedError("No factory method for type: " + type_)



import socket
import tempfile
from pathlib import Path

import rewrite.java.tree as j
import rewrite.python.tree as py
from rewrite import Markers
from rewrite import random_id, Cursor, PrintOutputCapture
from rewrite.java import Space, JavaType
from rewrite.python import Py
from rewrite.python.remote.receiver import PythonReceiver
from rewrite.python.remote.sender import PythonSender

from .receiver import ReceiverContext
from .remoting import (
    RemotingContext,
    RemotePrinterFactory,
)
from .sender import SenderContext

SenderContext.register(Py, lambda: PythonSender())
ReceiverContext.register(Py, lambda: PythonReceiver())

# Path to the Unix domain socket
SOCKET_PATH = tempfile.gettempdir() + "/rewrite-java.sock"

# Create a Unix domain socket
client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

# Connect the socket to the path where the server is listening
client.connect(("localhost", 65432))
print(f"Connected to {SOCKET_PATH}")

try:
    remoting = RemotingContext()
    remoting.connect(client)
    RemotePrinterFactory(remoting.client).set_current()

    literal = j.Literal(
        random_id(),
        Space.SINGLE_SPACE,
        Markers.EMPTY,
        True,
        "True",
        None,
        JavaType.Primitive(),
    )
    assert_ = py.AssertStatement(
        random_id(),
        Space.EMPTY,
        Markers.EMPTY,
        [j.JRightPadded(literal, Space.EMPTY, Markers.EMPTY)],
    )
    cu = py.CompilationUnit(
        random_id(),
        Space.EMPTY,
        Markers.EMPTY,
        Path("/foo.py"),
        None,
        None,
        False,
        None,
        [],
        [j.JRightPadded(assert_, Space.EMPTY, Markers.EMPTY)],
        Space.EMPTY,
    )
    printed = cu.print(Cursor(None, Cursor.ROOT_VALUE), PrintOutputCapture(0))
    assert printed == "assert True"
finally:
    client.close()



import importlib
import typing
from dataclasses import Field
from functools import lru_cache
from typing import Any, Type
from typing import Optional


def to_java_type_name_from_value(v: Any) -> Optional[str]:
    if isinstance(v, bool):
        return "java.lang.Boolean"
    elif isinstance(v, int):
        bit_length = v.bit_length()
        if bit_length < 32:
            return "java.lang.Integer"
        elif bit_length < 64:
            return "java.lang.Long"
        else:
            return "java.math.BigInteger"
    elif isinstance(v, complex):
        return "java.lang.String"
    return to_java_type_name(type(v))


@lru_cache(maxsize=None)
def to_java_type_name(t: typing.Type[Any]) -> Optional[str]:
    if t == bytes:
        # FIXME can we map this somehow?
        return None
    if t == bool:
        return "java.lang.Boolean"
    if t == int:
        return "java.lang.Integer"
    if t == str:
        return "java.lang.String"
    if t == float:
        return "java.lang.Double"
    if t == type(None):
        return "null"
    if t.__module__.startswith("rewrite.java.support_types"):
        if t.__name__ == "Space":
            return "org.openrewrite.java.tree.Space"
        if t.__name__ == "Comment":
            return "org.openrewrite.java.tree.Comment"
        if t.__name__ == "TextComment":
            return "org.openrewrite.java.tree.TextComment"
        if t.__name__ == "JLeftPadded":
            return "org.openrewrite.java.tree.JLeftPadded"
        if t.__name__ == "JRightPadded":
            return "org.openrewrite.java.tree.JRightPadded"
        if t.__name__ == "JContainer":
            return "org.openrewrite.java.tree.JContainer"
    if t.__module__.startswith("rewrite.java.markers"):
        return "org.openrewrite.java.marker." + t.__qualname__
    if t.__module__.startswith("rewrite.java.tree"):
        return "org.openrewrite.java.tree.J$" + t.__qualname__.replace(".", "$")
    if t.__module__.startswith("rewrite.python.support_types"):
        if t.__name__ == "PyComment":
            return "org.openrewrite.python.tree.PyComment"
        if t.__name__ == "PyLeftPadded":
            return "org.openrewrite.python.tree.PyLeftPadded"
        if t.__name__ == "PyRightPadded":
            return "org.openrewrite.python.tree.PyRightPadded"
        if t.__name__ == "PyContainer":
            return "org.openrewrite.python.tree.PyContainer"
    if t.__module__.startswith("rewrite.python.markers"):
        return "org.openrewrite.python.marker." + t.__qualname__
    if t.__module__.startswith("rewrite.python.style"):
        return "org.openrewrite.python.style." + t.__qualname__
    if t.__module__.startswith("rewrite.python.tree"):
        return "org.openrewrite.python.tree.Py$" + t.__qualname__.replace(".", "$")
    if t.__module__.startswith("rewrite.marker"):
        if t.__name__ == "ParseExceptionResult":
            return "org.openrewrite.ParseExceptionResult"
        return "org.openrewrite.marker." + t.__qualname__.replace(".", "$")
    if t.__module__ == "rewrite.parser" and t.__name__ == "ParseError":
        return "org.openrewrite.tree.ParseError"
    if t.__module__.startswith("rewrite.style"):
        return "org.openrewrite.style." + t.__qualname__
    if t.__module__.startswith("rewrite.") and t.__module__.endswith(".tree"):
        model = t.__module__.split(".")[1]
        return (
            "org.openrewrite."
            + model
            + ".tree."
            + model.capitalize()
            + "$"
            + t.__qualname__.replace(".", "$")
        )
    return t.__module__ + "." + t.__qualname__
    # raise NotImplementedError("to_java_type_name: " + str(o))


@lru_cache(maxsize=None)
def to_java_field_name(field: Field[Any]) -> str:
    return __convert_snake_to_camel(field.name[1:])


def __convert_snake_to_camel(field_name: str) -> str:
    # Remove leading underscore
    if field_name.startswith("_"):
        field_name = field_name[1:]

    # Convert snake case to camel case
    components = field_name.split("_")
    return components[0] + "".join(x.capitalize() for x in components[1:])


def get_type(type_name: str) -> Type[Any]:
    # `type_name` will look like `org.openrewrite.java.tree.J$CompilationUnit`
    parts = type_name.split('.')
    module = importlib.import_module('rewrite.' + parts[2])
    return getattr(module, parts[-1].split('$')[1])



import logging
import socket
from io import BytesIO
from typing import List, TypedDict
from cbor2 import dumps, CBORDecoder

from rewrite_remote.remote_utils import COMMAND_END
from rewrite_remote.remoting import OK, RemotingContext, RemotingMessageType
from rewrite_remote.handlers.handler_helpers import respond_with_error
from rewrite_remote.handlers.pypi_manager import PyPiManager, Source
from rewrite_remote.handlers.types import PackageSource


class RecipeInstallArgs(TypedDict):
    package_id: str
    package_version: str
    include_default_repository: bool
    package_sources: List[PackageSource]


def decode_recipe_install_args(decoder: CBORDecoder) -> RecipeInstallArgs:
    """
    Decodes the arguments (order matters and must match the order encoded)
    """
    package_id = str(decoder.decode())
    package_version = str(decoder.decode())
    include_default_repository = bool(decoder.decode())
    package_sources_data = decoder.decode()

    if not isinstance(package_sources_data, list):
        raise ValueError("package_sources_data is not a list")

    package_sources = []
    if package_sources_data and len(package_sources_data) > 0:
        package_sources = [
            PackageSource(source=ps.get("source"), credential=ps.get("credential"))
            for ps in package_sources_data
        ]

    return {
        "package_id": package_id,
        "package_version": package_version,
        "include_default_repository": include_default_repository,
        "package_sources": package_sources,
    }


# Main command handler with the specified signature
def recipe_install_handler(
    stream: BytesIO, sock: socket.socket, remoting_ctx: RemotingContext
) -> None:
    remoting_ctx.reset()

    # 1. Read input from stream
    try:
        data = stream.read()
        decoder = CBORDecoder(BytesIO(data))
        args = decode_recipe_install_args(decoder)
        package_id = args.get("package_id")
        package_version = args.get("package_version")
        include_default_repository = args.get("include_default_repository")
        package_sources = args.get("package_sources")
    except Exception as e:  # pylint: disable=broad-except
        respond_with_error(f"Failed to decode arguments: {e}", sock)
        return

    if package_id is None:
        respond_with_error("package_id is required", sock)
        return

    if package_version is None:
        respond_with_error("package_version is required", sock)
        return

    if package_sources is None:
        respond_with_error("package_sources is required", sock)
        return

    if include_default_repository is None:
        respond_with_error("include_default_repository is required", sock)
        return

    # 2. Log the request
    logging.info(
        f"""[Server] Handling install-recipe request: {{
        packageId: {package_id},
        packageVersion: {package_version},
        packageSources: {package_sources},
        includeDefaultRepository: {include_default_repository},
    }}"""
    )

    # 3. Validate sources
    sources: List[Source] = [
        Source(
            source=ps.source,
            username=ps.credential.get("username") if ps.credential else None,
            password=ps.credential.get("password") if ps.credential else None,
            token=ps.credential.get("token") if ps.credential else None,
        )
        for ps in package_sources or []
    ]

    valid_source = PyPiManager.find_valid_source(
        package_id, package_version, sources, include_default_repository
    )

    if not valid_source:
        respond_with_error("No valid sources found", sock)
        return

    # 4. Install the recipe
    try:
        installable_recipes = PyPiManager.install_package(package_id, package_version, valid_source)
    except Exception as e:  # pylint: disable=broad-except
        respond_with_error(f"Failed to install package: {e}", sock)
        return

    # 5. Log the result
    logging.info(
        "[Server] Found %d recipe(s) for package %s",
        len(installable_recipes.recipes),
        package_id,
    )
    for recipe in installable_recipes.recipes:
        logging.info("  Resolved recipe %s from %s", recipe, valid_source.source)

    # 6. Write response to stream
    response = {
        "recipes": [
            {
                "name": recipe.name,
                "source": recipe.source,
                "options": [],
            }
            for recipe in installable_recipes.recipes
        ],
        "repository": installable_recipes.source,
        "version": installable_recipes.version,
    }

    # Encode the response using CBOR
    encoded_response = b""
    encoded_response += dumps(RemotingMessageType.Response)
    encoded_response += dumps(OK)
    encoded_response += dumps(response)
    encoded_response += COMMAND_END
    sock.sendall(encoded_response)

    logging.info("[Server] Request completed.")



import logging
import socket
from io import BytesIO
from typing import List
from cbor2 import dumps, CBORDecoder

from rewrite_remote.handlers.project_helper import list_projects
from rewrite_remote.remote_utils import COMMAND_END
from rewrite_remote.remoting import OK, RemotingContext, RemotingMessageType
from rewrite_remote.handlers.handler_helpers import respond_with_error


# Main command handler with the specified signature
def list_projects_handler(
    stream: BytesIO, sock: socket.socket, remoting_ctx: RemotingContext
) -> None:
    remoting_ctx.reset()

    # 1. Read input from stream
    try:
        data = stream.read()
        decoder = CBORDecoder(BytesIO(data))
        root_project_file = str(decoder.decode())
    except Exception as e:  # pylint: disable=broad-except
        respond_with_error(f"Failed to decode arguments: {e}", sock)
        return

    if root_project_file == "" or root_project_file is None:
        respond_with_error("root_project_file is required", sock)
        return

    # 2. Log the request
    logging.info(
        f"""[Server] Handling install-recipe request: {{
        root_project_file: {root_project_file},
    }}"""
    )

    # 3. Find projects
    projects = list_projects(root_project_file)

    # 4. Log the result
    logging.info("[Server] Found %d project(s)", len(projects))
    for project in projects:
        logging.info(
            " %s root at %s using %s",
            project.project_name,
            project.project_root,
            project.project_tool,
        )

    # 5. Write response to stream
    response: List[str] = []

    for project in projects:
        response.append(project.project_root)

    # Encode the response using CBOR
    encoded_response = b""
    encoded_response += dumps(RemotingMessageType.Response)
    encoded_response += dumps(OK)
    encoded_response += dumps(response)
    encoded_response += COMMAND_END
    sock.sendall(encoded_response)

    logging.info("[Server] Request completed.")



# type: ignore
from __future__ import absolute_import

import decimal
from dataclasses import fields
from pathlib import Path
from typing import (
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Protocol,
    TYPE_CHECKING,
    Type,
    TypeVar,
)
from uuid import UUID

import cbor2
from cbor2 import CBOREncoder
from rewrite import Tree, Markers, Marker, ParseErrorVisitor, Cursor, Style
from rewrite.visitor import TreeVisitor

from rewrite_remote import remote_utils
from rewrite_remote.event import *
from rewrite_remote.remote_utils import Operation
from rewrite_remote.type_utils import (
    to_java_type_name,
    to_java_field_name,
    to_java_type_name_from_value,
)

if TYPE_CHECKING:
    from .remoting import RemotingContext

A = TypeVar("A")
T = TypeVar("T", bound=Tree)
V = TypeVar("V")
I = TypeVar("I")


class Sender(Protocol):
    def send(self, after: T, before: Optional[T], ctx: "SenderContext") -> None: ...


class OmniSender(Sender):
    def send(self, after: Tree, before: Optional[Tree], ctx: "SenderContext") -> None:
        sender = ctx.new_sender(type(after))
        sender.send(after, before, ctx)


class TreeSender(Protocol):
    def send_node(self, diff_event: DiffEvent, visitor: Callable[["TreeSender"], None]) -> None: ...

    def send_value(self, diff_event: DiffEvent) -> None: ...

    def flush(self) -> None: ...


class SenderContext(Generic[T]):
    """
    Context for serialization of tree nodes and values.
    """

    Registry: Dict[Type[Any], Callable[[], Sender]] = {}

    def __init__(
        self,
        sender: "TreeSender",
        visitor: TreeVisitor = None,
        before: Optional[Any] = None,
    ):
        self.sender = sender
        self.visitor = visitor
        self.before = before

    def new_sender(self, type_: Type[Any]) -> Sender:
        for entry_type, factory in self.Registry.items():
            # FIXME find better solution
            try:
                if type_.__bases__.__contains__(entry_type) or issubclass(type_, entry_type):
                    return factory()
            except:
                pass
        raise ValueError(f"Unsupported sender type: {type_}")

    def fork(self, visitor: TreeVisitor, before: Optional[Any] = None) -> "SenderContext[T]":
        return SenderContext(self.sender, visitor, before)

    def visit(
        self,
        consumer: Callable[[V, "SenderContext"], None],
        after: V,
        before: Optional[V] = None,
    ) -> None:
        save_before = self.before
        self.before = before
        consumer(after, self)
        self.before = save_before

    def send_tree(self, after: T, ctx: "SenderContext[T]") -> None:
        after.accept(self.visitor, ctx)

    def send_any_tree(self, after: T, before: Optional[T]) -> None:
        OmniSender().send(after, before, self)

    def send_node(
        self,
        owner: A,
        extractor: Callable[[A], Optional[V]],
        details: Callable[[V, "SenderContext[T]"], None],
    ) -> None:
        self.send_node_internal(
            extractor(owner),
            extractor(self.before) if self.before is not None else None,
            details,
        )

    def send_value(self, owner: Any, value_extractor: Callable[[T], Optional[V]]) -> None:
        after_value = value_extractor(owner)
        before_value = value_extractor(self.before) if self.before is not None else None
        self.send_value_internal(after_value, before_value)

    def send_typed_value(self, owner: A, value_extractor: Callable[[A], V]) -> None:
        after_value = value_extractor(owner)
        before_value = value_extractor(self.before) if self.before is not None else None
        self.send_typed_value_internal(after_value, before_value)

    def send_list_event(self, after: Optional[List[V]], before: Optional[List[V]]) -> bool:
        if after == before:
            evt = DiffEvent(EventType.NoChange, None, None)
        elif before is None:
            evt = DiffEvent(EventType.Add, None, len(after) if after is not None else 0)
        elif after is None:
            evt = DiffEvent(EventType.Delete, None, None)
        else:
            evt = DiffEvent(EventType.Update, None, len(after))

        self.sender.send_value(evt)
        return evt.event_type != EventType.NoChange and evt.event_type != EventType.Delete

    def send_nodes(
        self,
        owner: A,
        element_extractor: Callable[[A], List[V]],
        details: Callable[[V, "SenderContext"], None],
        id_function: Callable[[V], I],
    ) -> None:
        after_list = element_extractor(owner)
        before_list = element_extractor(self.before) if self.before is not None else None

        if self.send_list_event(after_list, before_list):
            if before_list is not None:
                self.sender.send_value(DiffEvent(EventType.StartList, None, None))

            remote_utils.calculate_list_diff(
                before_list or [],
                after_list,
                id_function,
                lambda op, _1, _2, bv, av: {
                    Operation.Delete: lambda: self.send_node_internal(av, bv, details),
                    Operation.NoChange: lambda: self.send_node_internal(av, bv, details),
                    Operation.Add: lambda: self.send_node_internal(av, bv, details),
                    Operation.Update: lambda: self.send_node_internal(av, bv, details),
                    Operation.Move: lambda: NotImplementedError("Unexpected operation: " + str(op)),
                }[op](),
            )

            if before_list is not None:
                self.sender.send_value(DiffEvent(EventType.EndList, None, None))

    def send_values(
        self,
        owner: T,
        value_extractor: Callable[[T], List[V]],
        id_function: Callable[[V], I],
    ) -> None:
        after_list = value_extractor(owner)
        before_list = value_extractor(self.before) if self.before is not None else None

        if self.send_list_event(after_list, before_list):
            if before_list is not None:
                self.sender.send_value(DiffEvent(EventType.StartList, None, None))

            remote_utils.calculate_list_diff(
                before_list or [],
                after_list,
                id_function,
                lambda op, _1, _2, bv, av: {
                    Operation.Delete: lambda: self.send_value_internal(av, bv),
                    Operation.NoChange: lambda: self.send_value_internal(av, bv),
                    Operation.Add: lambda: self.send_value_internal(av, bv),
                    Operation.Update: lambda: self.send_value_internal(av, bv),
                    Operation.Move: lambda: NotImplementedError("Unexpected operation: " + str(op)),
                }[op](),
            )

            if before_list is not None:
                self.sender.send_value(DiffEvent(EventType.EndList, None, None))

    def send_typed_values(
        self,
        owner: T,
        value_extractor: Callable[[T], List[V]],
        id_function: Callable[[V], I],
    ) -> None:
        after_list = value_extractor(owner)
        before_list = value_extractor(self.before) if self.before is not None else None

        if self.send_list_event(after_list, before_list):
            if before_list is not None:
                self.sender.send_value(DiffEvent(EventType.StartList, None, None))

            remote_utils.calculate_list_diff(
                before_list or [],
                after_list,
                id_function,
                lambda op, _1, _2, bv, av: {
                    Operation.Delete: lambda: self.send_typed_value_internal(av, bv),
                    Operation.NoChange: lambda: self.send_typed_value_internal(av, bv),
                    Operation.Add: lambda: self.send_typed_value_internal(av, bv),
                    Operation.Update: lambda: self.send_typed_value_internal(av, bv),
                    Operation.Move: lambda: NotImplementedError("Unexpected operation: " + str(op)),
                }[op](),
            )

            if before_list is not None:
                self.sender.send_value(DiffEvent(EventType.EndList, None, None))

    def send_markers(self, markers: Markers, ignore: Optional[bool]) -> None:
        self.send_value(markers, lambda ms: ms.id)
        self.send_values(markers, lambda ms: ms.markers, lambda ms: ms.id)

    def send_tree_visitor(self, after: T, ctx: "SenderContext") -> None:
        after.accept(self.visitor, ctx)

    @staticmethod
    def register(type_: Type, sender_factory: Callable[[], Sender]) -> None:
        SenderContext.Registry[type_] = sender_factory

    @staticmethod
    def are_equal(after: Optional[V], before: Optional[V]) -> bool:
        if after is None or before is None:
            return after == before

        return isinstance(after, (Tree, Markers)) or isinstance(before, (Tree, Markers))

    def send_node_internal(
        self,
        after: Optional[V],
        before: Optional[V],
        details: Callable[[V, "SenderContext"], None],
    ) -> None:
        if self.are_equal(after, before):
            evt = DiffEvent(EventType.NoChange, None, None)
        elif before is None:
            concrete_type = to_java_type_name(type(after)) if after is not None else None
            evt = DiffEvent(EventType.Add, concrete_type, None)
        elif after is None:
            evt = DiffEvent(EventType.Delete, None, None)
        else:
            evt = DiffEvent(EventType.Update, None, None)

        self.sender.send_node(evt, lambda _: self.visit(details, after, before))

    def send_value_internal(self, after: V, before: Optional[V]) -> None:
        if self.before is not None and self.are_equal(after, before):
            evt = DiffEvent(EventType.NoChange, None, None)
        elif self.before is None or before is None:
            concrete_type = to_java_type_name(type(after)) if isinstance(after, Marker) else None
            evt = DiffEvent(EventType.Add, concrete_type, after)
        elif after is None:
            evt = DiffEvent(EventType.Delete, None, None)
        else:
            evt = DiffEvent(EventType.Update, None, after)

        self.sender.send_value(evt)

    def send_typed_value_internal(self, after: V, before: Optional[V]) -> None:
        if self.before is not None and self.are_equal(after, before):
            evt = DiffEvent(EventType.NoChange, None, None)
        elif self.before is None or before is None:
            concrete_type = to_java_type_name_from_value(after) if after is not None else None
            evt = DiffEvent(EventType.Add, concrete_type, after)
        elif after is None:
            evt = DiffEvent(EventType.Delete, None, None)
        else:
            evt = DiffEvent(EventType.Update, None, after)

        self.sender.send_value(evt)


class SerializationContext:
    def __init__(
        self,
        remoting_context: "RemotingContext",
        value_serializers: Optional[Dict[Type[Any], "ValueSerializer"]] = None,
    ):
        self.remoting_context = remoting_context
        self.value_serializers = value_serializers or {}

    def serialize(self, value: Any, type_name: Optional[str], encoder: CBOREncoder) -> None:
        if value is None:
            encoder.encode_none(None)
            return

        for type_cls, serializer in self.value_serializers.items():
            if isinstance(value, type_cls):
                serializer(value, type_name, encoder, self)
                return

        DefaultValueSerializer().serialize(value, type_name, encoder, self)


class JsonSender(TreeSender):
    Debug = False

    def __init__(self, stream, context: SerializationContext):
        self._stream = stream
        self._context = context
        self._encoder = cbor2.CBOREncoder(self._stream)

    def send_node(self, diff_event, visitor) -> None:
        if diff_event.event_type in (EventType.Add, EventType.Update):
            self._encoder.encode(
                [diff_event.event_type.value]
                if diff_event.concrete_type is None
                else [diff_event.event_type.value, diff_event.concrete_type]
            )

            if self.Debug:
                print(f"SEND: {diff_event}")

            visitor(self)
        elif diff_event.event_type in (EventType.Delete, EventType.NoChange):
            self._encoder.encode([diff_event.event_type.value])

            if self.Debug:
                print(f"SEND: {diff_event}")
        else:
            raise NotImplementedError()

    def send_value(self, diff_event):
        if diff_event.event_type in (EventType.Add, EventType.Update):
            self._encoder.encode_length(4, 3 if diff_event.concrete_type is not None else 2)
            self._encoder.encode_int(diff_event.event_type.value)
            if diff_event.concrete_type is not None:
                self._encoder.encode(diff_event.concrete_type)
            self._context.serialize(diff_event.msg, diff_event.concrete_type, self._encoder)
        elif diff_event.event_type in (
            EventType.Delete,
            EventType.NoChange,
            EventType.StartList,
            EventType.EndList,
        ):
            self._encoder.encode([diff_event.event_type.value])
        else:
            raise NotImplementedError()

        if self.Debug:
            print(f"SEND: {diff_event}")

    def flush(self):
        self._stream.flush()


INDEFINITE_ARRAY_START = b"\x9f"
INDEFINITE_MAP_START = b"\xbf"
BREAK_MARKER = b"\xff"

ValueSerializer = Callable[[Any, Optional[str], CBOREncoder, SerializationContext], None]


def write_object_using_reflection(
    value: Any,
    type_name: Optional[str],
    with_id: bool,
    encoder: CBOREncoder,
    context: SerializationContext,
) -> None:
    if with_id and (id := context.remoting_context.try_get_id(value)):
        encoder.encode_int(id)
        return

    encoder.write(INDEFINITE_MAP_START)
    encoder.encode_string("@c")
    encoder.encode_string(type_name or to_java_type_name(type(value)))
    if with_id:
        encoder.encode_string("@ref")
        id = context.remoting_context.add(value)
        encoder.encode_int(id)

    for field in fields(value):
        if field.name[0] == "_" and (
            not hasattr(field.type, "__origin__") or field.type.__origin__ is not ClassVar
        ):
            encoder.encode_string(to_java_field_name(field))
            context.serialize(getattr(value, field.name), None, encoder)
    encoder.write(BREAK_MARKER)


class DefaultValueSerializer(ValueSerializer):
    def __call__(self, *args, **kwargs):
        return self.serialize(*args, **kwargs)

    def serialize(
        self,
        value: Any,
        type_name: Optional[str],
        encoder: CBOREncoder,
        context: SerializationContext,
    ) -> None:
        if isinstance(value, (int, float, str, bool, decimal.Decimal)):
            encoder.encode(value)
        elif value is None:
            encoder.encode_none(None)
        elif isinstance(value, UUID):
            encoder.encode(value.bytes)
        elif isinstance(value, Enum):
            encoder.encode(value.value)
        elif isinstance(value, Path):
            encoder.encode(str(value))
        elif isinstance(value, (list, set, tuple)):
            encoder.encode_length(4, len(value))
            for item in value:
                context.serialize(item, None, encoder)
        elif isinstance(value, Markers):
            if id := context.remoting_context.try_get_id(value):
                encoder.encode_int(id)
            else:
                id = context.remoting_context.add(value)
                encoder.encode_length(5, 3)
                encoder.encode_string("@ref")
                encoder.encode_int(id)
                encoder.encode_string("id")
                encoder.encode_uuid(value.id)
                encoder.encode_string("markers")
                encoder.encode_length(4, len(value.markers))
                for marker in value.markers:
                    context.serialize(marker, None, encoder)
        elif isinstance(value, (Marker, Style)):
            if id := context.remoting_context.try_get_id(value):
                encoder.encode_int(id)
            else:
                id = context.remoting_context.add(value)
                encoder.write(INDEFINITE_MAP_START)
                encoder.encode_string("@c")
                encoder.encode_string(to_java_type_name(type(value)))
                encoder.encode_string("@ref")
                encoder.encode_int(id)
                for field in fields(value):
                    if field.name[0] == "_" and (
                        not hasattr(field.type, "__origin__")
                        or field.type.__origin__ is not ClassVar
                    ):
                        encoder.encode_string(to_java_field_name(field))
                        context.serialize(getattr(value, field.name), None, encoder)
                encoder.write(BREAK_MARKER)
        elif isinstance(value, bytes):
            # FIXME verify that this can be deserialized
            encoder.encode(value)
        elif isinstance(value, complex):
            encoder.encode(str(value))
        else:
            write_object_using_reflection(value, type_name, False, encoder, context)


def delegate_based_serializer(
    delegate: Callable[[Any, Optional[str], CBOREncoder, SerializationContext], None],
) -> Type[ValueSerializer]:
    class DelegateBasedSerializer(ValueSerializer):
        def serialize(
            self,
            value: Any,
            type_name: Optional[str],
            encoder: CBOREncoder,
            context: SerializationContext,
        ) -> None:
            delegate(value, type_name, encoder, context)

    return DelegateBasedSerializer


class ParseErrorSender(Sender):
    def send(self, after: Tree, before: Optional[Tree], ctx: "SenderContext") -> None:
        visitor = self.Visitor()
        visitor.visit(after, ctx.fork(visitor, before))

    class Visitor(ParseErrorVisitor):
        def visit(self, tree, ctx, parent: Optional[Cursor] = None):
            self.cursor = Cursor(self.cursor, tree)
            ctx.send_node(tree, lambda x: x, ctx.send_tree)
            self.cursor = self.cursor.parent

            return tree

        def visit_parse_error(self, parse_error, ctx):
            ctx.send_value(parse_error, lambda v: v.id)
            ctx.send_node(parse_error, lambda v: v.markers, ctx.send_markers)
            ctx.send_value(parse_error, lambda v: v.source_path)
            ctx.send_typed_value(parse_error, lambda v: v.file_attributes)
            ctx.send_value(parse_error, lambda v: v.charset_name)
            ctx.send_value(parse_error, lambda v: v.charset_bom_marked)
            ctx.send_typed_value(parse_error, lambda v: v.checksum)
            ctx.send_value(parse_error, lambda v: v.text)
            # ctx.send_node(parse_error, lambda v: v.erroneous, ctx.send_tree)
            return parse_error



import logging
import socket
from io import BytesIO
from typing import List, TypedDict, Any
from cbor2 import dumps, CBORDecoder

from rewrite_remote.handlers.pypi_manager import Option, PyPiManager
from rewrite_remote.remoting import (
    OK,
    RemotingExecutionContextView,
    RemotingMessageType,
    RemotingMessenger,
)

from rewrite_remote.remote_utils import COMMAND_END
from rewrite_remote.remoting import RemotingContext
from rewrite_remote.handlers.handler_helpers import respond_with_error

from rewrite import InMemoryExecutionContext


class RunRecipeLoadAndVisitorArgs(TypedDict):
    recipe_name: str
    recipe_source: str
    recipe_options: Any  # List[recipeOption]


def decode_run_recipe_load_and_visitor_args(
    decoder: CBORDecoder,
) -> RunRecipeLoadAndVisitorArgs:
    """
    Decodes the arguments (order matters and must match the order encoded)
    """
    recipe_name = str(decoder.decode())
    recipe_source = str(decoder.decode())
    recipe_options = decoder.decode()

    return {
        "recipe_name": recipe_name,
        "recipe_source": recipe_source,
        "recipe_options": recipe_options,
    }


def run_recipe_load_and_visitor_handler(
    stream: BytesIO, sock: socket.socket, remoting_ctx: RemotingContext
) -> None:
    remoting_ctx.reset()

    # Read input from stream
    try:
        data = stream.read()
        decoder = CBORDecoder(BytesIO(data))
        args = decode_run_recipe_load_and_visitor_args(decoder)
        recipe_name = args.get("recipe_name")
        recipe_source = args.get("recipe_source")
        recipe_options: List[Option] = args.get("recipe_options") or []
    except Exception as e:  # pylint: disable=broad-except
        respond_with_error(f"Failed to decode arguments: {e}", sock)
        return

    if recipe_name is None:
        respond_with_error("recipe_name is required", sock)
        return

    if recipe_source is None:
        respond_with_error("recipe_source is required", sock)
        return

    # Log the request
    logging.info(
        f"""[Server] Handling run-recipe-load-and-visitor request: {{
        recipe_name: {recipe_name},
        recipe_source: {recipe_source},
        recipe_options: {recipe_options},
    }}"""
    )

    # Receive the tree
    if not hasattr(RemotingMessenger, "_state"):
        RemotingMessenger._state = None

    if not hasattr(RemotingMessenger, "_context"):
        RemotingMessenger._context = remoting_ctx

    received = None
    try:
        received = RemotingMessenger.receive_tree(
            RemotingMessenger, stream, RemotingMessenger._state
        )
    except Exception as e:
        # logging.error(f"Failed to receive tree: {e}")
        respond_with_error(f"Failed to receive tree: {e}", sock)
        return

    # Set the execution context
    ctx = InMemoryExecutionContext()
    RemotingExecutionContextView.view(ctx).remoting_context = remoting_ctx

    try:
        recipe_instance = PyPiManager.load_recipe(recipe_name, recipe_source, recipe_options)
    except Exception as e:
        respond_with_error(f"Failed to load recipe: {e}", sock)
        return

    # 4. Run the recipe
    try:
        tree_visitor = recipe_instance.get_visitor()

        if not hasattr(tree_visitor, "visit") or not callable(tree_visitor.visit):
            raise ValueError("Visitor does not have a visit method")

        RemotingMessenger._state = tree_visitor.visit(received, ctx)

        if RemotingMessenger._state is None:
            raise ValueError("RemotingMessenger.state cannot be None")
    except Exception as e:
        respond_with_error(f"Failed to process input data with recipe: {e}", sock)
        return

    # 5. Write the response
    response_encoder = BytesIO()
    RemotingMessenger.send_tree(remoting_ctx, response_encoder, RemotingMessenger._state, received)

    encoded_response = b""
    encoded_response += dumps(RemotingMessageType.Response)
    encoded_response += dumps(OK)
    encoded_response += COMMAND_END
    sock.sendall(encoded_response)

    logging.info("[Server] Request completed.")



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any


class EventType(Enum):
    NoChange = 0
    Update = 1
    Add = 2
    Delete = 3
    Move = 4
    StartList = 5
    EndList = 6


@dataclass(frozen=True)
class DiffEvent:
    event_type: EventType
    concrete_type: Optional[str] = None
    msg: Optional[Any] = None



# type: ignore
import importlib
import importlib.resources
import logging
import os
import select
import socket
import sys
import time
import traceback
import zipfile
from io import BytesIO, StringIO

import cbor2
from cbor2 import dumps
from rewrite_remote.handlers.hello_world_handler import hello_world_handler
from rewrite_remote.handlers.list_projects_handler import list_projects_handler
from rewrite_remote.handlers.parse_project_sources_handler import (
    parse_project_sources_handler,
)
from rewrite_remote.handlers.project_helper import read_file_contents
from rewrite_remote.handlers.recipe_install_handler import (
    recipe_install_handler,
)
from rewrite_remote.handlers.run_recipe_load_and_visitor_handler import (
    run_recipe_load_and_visitor_handler,
)
from rewrite_remote.remoting import (
    RemotePrinterFactory,
    RemotingContext,
    RemotingMessageType,
    RemotingMessenger,
)
from rewrite_remote.sender import ParseErrorSender

from rewrite import (
    ParserInput,
    InMemoryExecutionContext,
    ExecutionContext,
    ParseError,
    Recipe,
)
from rewrite.java.remote import *
from rewrite.python import Py
from rewrite.python.parser import PythonParserBuilder
from rewrite.python.remote.receiver import PythonReceiver
from rewrite.python.remote.sender import PythonSender

INACTIVITY_TIMEOUT = 300  # 5 minutes
_OK: int = 0
_ERROR: int = 1


# Configure logging
logging.basicConfig(
    level=logging.DEBUG,
    format='%(asctime)s - %(levelname)s - %(message)s',
)
logger = logging.getLogger('server')
logger.setLevel(logging.DEBUG)

def register_remoting_factories() -> None:
    SenderContext.register(ParseError, ParseErrorSender)
    SenderContext.register(Py, PythonSender)
    ReceiverContext.register(Py, PythonReceiver)
    SenderContext.register(J, JavaSender)
    ReceiverContext.register(J, JavaReceiver)


def find_free_port() -> Any:
    """Find a free port by using the system to allocate a port for us."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        s.listen(1)
        return s.getsockname()[1]


class Server:
    """
    A server that listens for incoming connections and processes requests from clients.
    """

    _port: Optional[int]
    _path: Optional[str]
    _remoting_context: RemotingContext
    _messenger: RemotingMessenger

    def __init__(
        self,
        port: Optional[int] = None,
        path: Optional[str] = None,
        timeout: int = INACTIVITY_TIMEOUT,
    ):
        self._port = port
        self._path = path
        self.timeout = timeout
        self._remoting_context = RemotingContext()
        self._remoting_context._recipe_factories["test"] = lambda recipe_options: Recipe()
        self._messenger = RemotingMessenger(
            self._remoting_context,
            {
                "parse-python-source": self.parse_python_source,
                "parse-python-file": self.parse_python_file,
                "hello-world": hello_world_handler,
                "recipe-install": recipe_install_handler,
                "run-recipe-load-and-visitor": run_recipe_load_and_visitor_handler,
                "list-projects": list_projects_handler,
                "parse-project-sources": parse_project_sources_handler,
            },
        )

    def start(self) -> None:
        """Start the server and listen for connections on the given port."""
        if self._path:
            if os.path.exists(self._path):
                os.remove(self._path)
            with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
                s.bind(self._path)
                s.listen()
                logger.info(f"Server listening on Unix domain socket: {self._path}")
                while True:
                    conn, _ = s.accept()
                    with conn:
                        self.handle_client(conn)
        else:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                s.bind(("localhost", self._port))
                s.listen(5)
                logger.info(f"Server listening on port {self._port}")
                last_activity_time = time.time()
                while True:
                    s.settimeout(5)
                    try:
                        conn, addr = s.accept()
                        last_activity_time = time.time()  # Reset inactivity timer
                        with conn:
                            conn.settimeout(None)
                            self.handle_client(conn)
                    except socket.timeout:
                        current_time = time.time()
                        if current_time - last_activity_time >= self.timeout:
                            logger.info("No new connections for 5 minutes, shutting down server.")
                            break

    def handle_client(self, sock: socket.socket) -> None:
        try:
            self._remoting_context.connect(sock)
            RemotePrinterFactory(self._remoting_context.client).set_current()

            sock.setblocking(True)
            while True:
                message_type = sock.recv(1)
                if not message_type:
                    return
                assert cbor2.load(BytesIO(message_type)) == RemotingMessageType.Request
                self._messenger.process_request(sock)

        except (OSError, IOError):
            logger.error("Socket was closed unexpectedly")
            return
        except Exception as e:
            logger.error(f"An error occurred while handling client: {e}")
            traceback.print_exc()
            if sock.fileno() != -1:
                try:
                    # Equivalent to C#'s stream.WriteTimeout = 1000;
                    sock.send(dumps(RemotingMessageType.Response))
                    sock.send(dumps(_ERROR))
                    sock.send(dumps(traceback.format_exc()))
                except (OSError, IOError):
                    logger.error("Failed to send error response, socket was closed")
                    return
                except Exception as inner_exception:
                    logger.error(f"An error occurred while sending error response: {inner_exception}")

    def parse_python_source(
        self,
        stream: BytesIO,
        sock: socket.socket,
        remoting_ctx: RemotingContext,
    ) -> None:
        remoting_ctx.reset()
        source = cbor2.load(stream)
        ctx = InMemoryExecutionContext()
        ctx.put_message(ExecutionContext.REQUIRE_PRINT_EQUALS_INPUT, False)
        for cu in (
            PythonParserBuilder()
            .build()
            .parse_inputs(
                [ParserInput(Path("source.py"), None, True, lambda: StringIO(source))],
                None,
                ctx,
            )
        ):
            response_stream = BytesIO()
            cbor2.dump(RemotingMessageType.Response, response_stream)
            cbor2.dump(_OK, response_stream)
            source_stream = BytesIO()
            remoting_ctx.new_sender_context(source_stream).send_any_tree(cu, None)
            cbor2.dump(source_stream.getvalue(), response_stream)
            sock.sendall(response_stream.getvalue())

    def parse_python_file(
        self,
        stream: BytesIO,
        sock: socket.socket,
        remoting_ctx: RemotingContext,
    ) -> None:
        remoting_ctx.reset()
        path = cbor2.load(stream)
        ctx = InMemoryExecutionContext()
        ctx.put_message(ExecutionContext.REQUIRE_PRINT_EQUALS_INPUT, False)
        for cu in (
            PythonParserBuilder()
            .build()
            .parse_inputs(
                [
                    ParserInput(
                        Path(path),
                        None,
                        True,
                        lambda: read_file_contents(path),
                    )
                ],
                None,
                ctx,
            )
        ):
            response_stream = BytesIO()
            cbor2.dump(RemotingMessageType.Response, response_stream)
            cbor2.dump(_OK, response_stream)
            source_stream = BytesIO()
            remoting_ctx.new_sender_context(source_stream).send_any_tree(cu, None)
            cbor2.dump(source_stream.getvalue(), response_stream)
            sock.sendall(response_stream.getvalue())


def read_data_from_zip() -> None:
    # Access the resource within the 'your_package.resources' package
    # 'data.zip' is the name of the file included
    with importlib.resources.open_binary("resources", "rewrite-remote-java.zip") as f:
        # Open the zip file from the resource file stream
        with zipfile.ZipFile(f) as zip_file:
            # List the contents of the zip file
            print(zip_file.namelist())
            # Read a specific file inside the zip file (if you know the file name within it)
            with zip_file.open(
                "rewrite-remote-java-0.2.0-SNAPSHOT/bin/rewrite-remote-java"
            ) as inner_file:
                data = inner_file.read()
                print(data.decode("utf-8"))


def main() -> Any:
    sys.setrecursionlimit(2000)
    port = int(sys.argv[1]) if len(sys.argv) > 1 else 54322
    timeout = int(sys.argv[2]) if len(sys.argv) > 2 else INACTIVITY_TIMEOUT
    register_remoting_factories()
    Server(port=port, timeout=timeout).start()
    # Server(port=find_free_port()).start()
    # Server(path=tempfile.gettempdir() + '/rewrite-csharp.sock').start()


if __name__ == "__main__":
    main()



# type: ignore
from __future__ import absolute_import

import socket
import threading
import traceback
from dataclasses import dataclass
from io import BytesIO
from threading import Lock
from typing import (
    Any,
    Dict,
    Optional,
    Type,
    Callable,
    cast,
    Union,
    BinaryIO,
)

import cbor2
from cbor2 import dumps, loads, load

from rewrite import (
    Recipe,
    InMemoryExecutionContext,
    Cursor,
    PrinterFactory,
    TreeVisitor,
    Tree,
)
from rewrite.execution import DelegatingExecutionContext
from rewrite.tree import PrintOutputCapture, P
from rewrite.visitor import T

from rewrite_remote import (
    ValueSerializer,
    ValueDeserializer,
    SenderContext,
    ReceiverContext,
    JsonSender,
    JsonReceiver,
    SerializationContext,
    DeserializationContext,
    remote_utils,
)


class RemotingContext:
    """
    A context for remoting operations.
    """

    _remoting_thread_local = threading.local()
    _recipe_factories: Dict[str, Callable[[str, Dict[str, Any]], Recipe]] = {}
    _value_serializers: Dict[Type, ValueSerializer] = {}
    _value_deserializers: Dict[str, ValueDeserializer] = {}
    _object_to_id_map: Dict[int, int] = {}
    _id_to_object_map: Dict[int, Any] = {}

    def __init__(self):
        self._client = None

    @classmethod
    def current(cls) -> "RemotingContext":
        result = getattr(cls._remoting_thread_local, "context", None)
        if result is None:
            raise ValueError("No RemotingContext has been set")
        return result

    def set_current(self) -> None:
        cls = self.__class__
        cls._remoting_thread_local.context = self

    @property
    def client(self) -> Optional["RemotingClient"]:
        return self._client

    def connect(self, sock: Any) -> "RemotingContext":
        self._client = RemotingClient(self, sock)
        return self

    def close(self) -> None:
        self._client.close()

    def try_get_id(self, key: Any) -> Optional[int]:
        return self._object_to_id_map.get(id(key))

    def add(self, value: Any) -> int:
        object_id = len(self._object_to_id_map)
        self._object_to_id_map[id(value)] = object_id
        self._id_to_object_map[object_id] = value
        return object_id

    def add_by_id(self, key: int, value: Any) -> None:
        self._id_to_object_map[key] = value
        self._object_to_id_map[id(value)] = key

    def get_object(self, key: int) -> Optional[Any]:
        return self._id_to_object_map.get(key)

    def reset(self) -> None:
        self._object_to_id_map.clear()
        self._id_to_object_map.clear()

    def new_sender_context(self, output_stream: Any) -> "SenderContext":
        return SenderContext(
            JsonSender(
                output_stream,
                SerializationContext(self, self._value_serializers),
            )
        )

    def new_receiver_context(self, input_stream: Any) -> "ReceiverContext":
        return ReceiverContext(
            JsonReceiver(
                input_stream,
                DeserializationContext(self, self._value_deserializers),
            )
        )

    def copy(self) -> "RemotingContext":
        return RemotingContext()

    def new_recipe(self, recipe_id: str, recipe_options: Any) -> "Recipe":
        return self._recipe_factories[recipe_id](recipe_options)

    @classmethod
    def register_value_serializer(cls, type_: Type, serializer: ValueSerializer) -> None:
        cls._value_serializers[type_] = serializer

    @classmethod
    def register_value_deserializer(cls, type_name: str, deserializer: ValueDeserializer) -> None:
        cls._value_deserializers[type_name] = deserializer


class RemotingExecutionContextView(DelegatingExecutionContext):
    def __init__(self, delegate):
        super().__init__(delegate)
        self._delegate = delegate

    @staticmethod
    def view(ctx):
        if isinstance(ctx, RemotingExecutionContextView):
            return ctx
        return RemotingExecutionContextView(ctx)

    @property
    def remoting_context(self) -> RemotingContext:
        return self._delegate.get_message("remoting", RemotingContext.current())

    @remoting_context.setter
    def remoting_context(self, value: RemotingContext):
        value.set_current()
        self._delegate.put_message("remoting", value)


OK = 0
ERROR = 1


class RemotingMessenger:
    def __init__(
        self,
        context: RemotingContext,
        additional_handlers: Dict[
            str, Callable[[BytesIO, socket.socket, RemotingContext], Any]
        ] = None,
    ):
        self._context = context
        self._additional_handlers = additional_handlers or {}
        self._recipes = []
        self._state = None

    def process_request(self, sock: socket.socket) -> bool:
        stream = remote_utils.read_to_command_end(sock)
        command = cbor2.load(stream)

        try:
            if command == "hello":
                self.handle_hello_command(stream, sock)
            elif command == "reset":
                self.handle_reset_command(stream, sock)
            elif command == "load-recipe":
                self.handle_load_recipe_command(stream, sock)
            elif command == "run-recipe-visitor":
                self.handle_run_recipe_visitor_command(stream, sock)
            elif command == "print":
                self.handle_print_command(stream, sock)
            else:
                if command in self._additional_handlers:
                    self._additional_handlers[command](stream, sock, self._context)
                else:
                    raise NotImplementedError(f"Unsupported command: {command}")
        except:
            traceback.print_exc()
        finally:
            self.send_end_message(sock)

        return True

    def handle_hello_command(self, stream: BytesIO, sock: socket.socket):
        cbor2.load(stream)
        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(OK, response_stream)
        sock.sendall(response_stream.getvalue())

    def handle_reset_command(self, stream: BytesIO, sock: socket.socket):
        self._state = None
        self._context = self._context.copy()
        self._context.connect(socket.socket())
        self._recipes.clear()
        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(OK, response_stream)
        sock.sendall(response_stream.getvalue())

    def handle_load_recipe_command(self, stream: BytesIO, sock: socket.socket):
        recipe_id = cbor2.load(stream)
        recipe_options = cbor2.load(stream)
        recipe = self._context.new_recipe(recipe_id, recipe_options)
        self._recipes.append(recipe)

        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(OK, response_stream)
        cbor2.dump(len(self._recipes) - 1, response_stream)
        sock.sendall(response_stream.getvalue())

    def handle_run_recipe_visitor_command(self, stream: BytesIO, sock: socket.socket):
        recipe_index = cbor2.load(stream)
        recipe = self._recipes[recipe_index]
        received = self.receive_tree(stream, self._state)

        ctx = InMemoryExecutionContext()
        RemotingExecutionContextView.view(ctx).remoting_context = self._context
        self._state = recipe.get_visitor().visit(received, ctx)

        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(OK, response_stream)
        self.send_tree(response_stream, self._state, received)
        sock.sendall(response_stream.getvalue())

    def handle_print_command(self, stream: BytesIO, sock: socket.socket):
        received = self.receive_tree(sock, None)
        root_cursor = Cursor(None, Cursor.ROOT_VALUE)
        ctx = InMemoryExecutionContext()
        RemotingExecutionContextView.view(ctx).remoting_context = self._context
        print_output = received.print(Cursor(root_cursor, received), PrintOutputCapture(0))

        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(OK, response_stream)
        cbor2.dump(print_output, response_stream)
        sock.sendall(response_stream.getvalue())

    def send_request(self, sock: socket.socket, command: str, *args):
        sock.sendall(dumps(RemotingMessageType.Request))
        sock.sendall(dumps(command))
        for arg in args:
            # FIXME serialize properly
            sock.sendall(dumps(arg))
        self.send_end_message(sock)

    def __send_request_stream(self, sock: socket.socket, command: str, *args):
        sock.sendall(dumps(RemotingMessageType.Request))
        sock.sendall(dumps(command))
        for arg in args:
            arg(sock)
        self.send_end_message(sock)

    @staticmethod
    def send_end_message(sock):
        sock.sendall(b"\x81\x17")

    def send_print_request(self, sock: socket.socket, cursor: Cursor):
        self.__send_request_stream(
            sock,
            "print",
            lambda s: self.send_tree(s, cast(Tree, cursor.value)),
        )
        if self.recv_byte(sock) != RemotingMessageType.Response:
            raise ValueError("Unexpected message type.")
        if self.recv_byte(sock) != 0:
            raise ValueError(f"Remote print failed: {loads(self.recv_all(sock))}")
        data = remote_utils.read_to_command_end(sock)
        print_msg = load(data)
        # end = load(data)  # end
        return print_msg

    def send_tree(
        self,
        dest: Union[BinaryIO, socket.socket],
        after: Tree,
        before: Optional[Tree] = None,
    ):
        b = BytesIO()
        self._context.new_sender_context(b).send_any_tree(after, before)
        if isinstance(dest, socket.socket):
            dest.sendall(dumps(b.getvalue()))
        else:
            b.seek(0)
            dest.write(dumps(b.getvalue()))

    def receive_tree(
        self,
        data: Union[BinaryIO, socket.socket],
        before: Optional[Tree] = None,
    ):
        receiver_context = self._context.new_receiver_context(BytesIO(cbor2.load(data)))
        return receiver_context.receive_any_tree(before)

    def send_run_recipe_request(
        self, sock: socket.socket, recipe, options: dict, source_files: list
    ):
        self.__send_request_stream(
            sock,
            "run-recipe",
            lambda s: (
                sock.sendall(dumps(recipe)),
                sock.sendall(dumps(options)),
                sock.sendall(dumps(len(source_files))),
                *[self.send_tree(sock, sf, None) for sf in source_files],
            ),
        )
        while self.recv_byte(sock) == RemotingMessageType.Request:
            self.process_request(sock)
        if self.recv_byte(sock) != RemotingMessageType.Response:
            raise ValueError("Unexpected message type.")
        if self.recv_byte(sock) != 0:
            raise ValueError(f"Remote recipe run failed: {loads(self.recv_all(sock))}")
        input_stream = remote_utils.read_to_command_end(sock)
        updated = [self.receive_tree(input_stream, sf) for sf in source_files]
        loads(input_stream)  # end
        return updated

    def send_reset_request(self, sock: socket.socket):
        self.send_request(sock, "reset")
        if self.recv_byte(sock) != RemotingMessageType.Response:
            raise ValueError("Unexpected message type.")
        if self.recv_byte(sock) != 0:
            raise ValueError(f"Remote reset failed: {loads(self.recv_all(sock))}")
        loads(self.recv_all(sock))  # command end

    def recv_byte(self, sock):
        return sock.recv(1)[0]

    def recv_all(self, sock, buffer_size=4096):
        data = b""
        while True:
            part = sock.recv(buffer_size)
            data += part
            if len(part) < buffer_size:
                break
        return data


class RemotingMessageType:
    Request = 0
    Response = 1


class RemotingClient:
    def __init__(self, context, sock: socket.socket):
        self._messenger = RemotingMessenger(context)
        self._socket = sock
        self._lock = Lock()

    def close(self) -> None:
        self._socket.close()

    def hello(self) -> None:
        with self._lock:
            self._messenger.send_request(self._socket, "hello")

    def print(self, cursor: Cursor) -> Any:
        with self._lock:
            return self._messenger.send_print_request(self._socket, cursor)

    def reset(self) -> None:
        with self._lock:
            self._messenger.send_reset_request(self._socket)

    def run_recipe(self, recipe, options: dict, source_files: list):
        with self._lock:
            return self._messenger.send_run_recipe_request(
                self._socket, recipe, options, source_files
            )


@dataclass
class RemotePrinter(TreeVisitor[Any, PrintOutputCapture[P]]):
    _client: RemotingClient

    def visit(
        self,
        tree: Optional[Tree],
        p: PrintOutputCapture[P],
        parent: Optional[Cursor] = None,
    ) -> Optional[T]:
        self.cursor = Cursor(parent, tree)
        p.append(self._client.print(self.cursor))
        self.cursor = self.cursor.parent
        return tree


@dataclass
class RemotePrinterFactory(PrinterFactory):
    _client: RemotingClient

    def create_printer(self, cursor: Cursor) -> TreeVisitor[Any, PrintOutputCapture[P]]:
        return RemotePrinter(self._client)



__path__ = __import__("pkgutil").extend_path(__path__, __name__)
from typing import TypeVar

from .receiver import *
from .sender import *
from .remoting import *

__all__ = [
    name for name in dir() if not name.startswith("_") and not isinstance(globals()[name], TypeVar)
]



import logging
import socket
import cbor2

from rewrite_remote.remote_utils import COMMAND_END
from rewrite_remote import RemotingMessageType, ERROR


def respond_with_error(message: str, sock: socket.socket) -> None:
    logging.error(f"[Server] Error: {message}")
    encoded_message = b""
    encoded_message += cbor2.dumps(RemotingMessageType.Response)
    encoded_message += cbor2.dumps(ERROR)
    encoded_message += cbor2.dumps(message)
    encoded_message += COMMAND_END
    sock.sendall(encoded_message)



import os
import glob
import toml
from typing import Any, List, Iterable
from pathlib import Path
from io import StringIO

from rewrite.tree import SourceFile
from rewrite.python.parser import PythonParserBuilder

from rewrite import ParserInput, InMemoryExecutionContext

from dataclasses import dataclass


@dataclass
class Project:
    project_name: str
    project_root: str
    project_tool: str


def list_projects(pyproject_path: str) -> list[Project]:
    """
    Parses the pyproject.toml file to identify sub projects in a monorepo and returns a list of Projects.
    """
    if not os.path.isfile(pyproject_path):
        raise FileNotFoundError(f"{pyproject_path} does not exist.")

    # Load and parse the pyproject.toml file
    with open(pyproject_path, "r") as file:
        data = toml.load(file)

    sub_projects: list[Project] = []

    if is_poetry_project(data):
        sub_projects = find_sub_projects_in_poetry(data, pyproject_path)
    elif is_hatch_project(data):
        sub_projects = find_sub_projects_in_hatch(data, pyproject_path)
    elif is_uv_project(data):
        sub_projects = find_sub_projects_in_uv_sources(data, pyproject_path)
    else:
        sub_projects = find_sub_projects_in_project_dependencies(data, pyproject_path)

    return sub_projects


def is_poetry_project(tomlData: dict[str, Any]) -> bool:
    return (
        "tool" in tomlData
        and "poetry" in tomlData["tool"]
        and "dependencies" in tomlData["tool"]["poetry"]
    )


def is_hatch_project(tomlData: dict[str, Any]) -> bool:
    return "tool" in tomlData and "hatch" in tomlData["tool"]


def is_uv_project(tomlData: dict[str, Any]) -> bool:
    return "tool" in tomlData and "uv" in tomlData["tool"]


def find_sub_projects_in_poetry(tomlData: dict[str, Any], toml_path: str) -> list[Project]:
    """
    Finds sub projects in a poetry project by looking for dependencies with a "path" key:
        [tool.poetry.dependencies]
        python = "^3.9"
        service-a = { path = "./services/service-a" }
        service-b = { path = "./services/service-b" }
        shared-lib = { path = "./shared-libraries/shared-lib" }
    """
    subProjects: list[Project] = []
    for dep_name, dep_value in tomlData["tool"]["poetry"]["dependencies"].items():
        if isinstance(dep_value, dict) and "path" in dep_value:
            subProjects.append(
                Project(
                    project_name=dep_name,
                    project_root=get_absolute_path(toml_path, dep_value["path"]),
                    project_tool="poetry",
                )
            )
    return subProjects


def find_sub_projects_in_hatch(tomlData: dict[str, Any], toml_path: str) -> list[Project]:
    """
    Finds sub projects in a hatch project by looking for dependencies with a "path" key:
        [tool.hatch.envs.default.dependencies]
        service-a = { path = "./services/service-a" }
        service-b = { path = "./services/service-b" }
    """
    subProjects: list[Project] = []
    hatch_envs = tomlData["tool"]["hatch"].get("envs", {})
    for env_name, env_data in hatch_envs.items():
        if isinstance(env_data, dict) and "dependencies" in env_data:
            dependencies = env_data["dependencies"]
            if isinstance(dependencies, dict):
                for dep_name, dep_value in dependencies.items():
                    if isinstance(dep_value, dict) and "path" in dep_value:
                        subProjects.append(
                            Project(
                                project_name=dep_name,
                                project_root=get_absolute_path(toml_path, dep_value["path"]),
                                project_tool=f"hatch:{env_name}",
                            )
                        )
    return subProjects


def find_sub_projects_in_uv_sources(tomlData: dict[str, Any], toml_path: str) -> list[Project]:
    """
    Finds sub projects in a uv project by looking at sources and workspace:
    [tool.uv.sources]
    service-a = { path = "./services/service-a" }
    service-b = { path = "./services/service-b" }

    [tool.uv.workspace]
    members = ["packages/*"]
    exclude = ["packages/excluded/*"]
    """
    subProjects: list[Project] = []
    uv_sources = tomlData["tool"]["uv"].get("sources", {})
    uv_workspace = tomlData["tool"]["uv"].get("workspace", {})
    for source_name, source_data in uv_sources.items():
        if isinstance(source_data, dict) and "path" in source_data:
            subProjects.append(
                Project(
                    project_name=source_name,
                    project_root=get_absolute_path(toml_path, source_data["path"]),
                    project_tool="uv.sources",
                )
            )
    if isinstance(uv_workspace, dict) and "members" in uv_workspace:
        excluded_directories = []
        exclude_globs = uv_workspace.get("exclude", [])
        for exclude_glob in exclude_globs:
            excluded_directories.extend(
                glob.glob(
                    os.path.join(os.path.dirname(toml_path), exclude_glob),
                    recursive=True,
                )
            )

        for glob_pattern in uv_workspace["members"]:
            # Every directory included by the members globs (and not excluded by the exclude globs) must contain a pyproject.toml file
            directories = glob.glob(
                os.path.join(os.path.dirname(toml_path), glob_pattern),
                recursive=True,
            )

            for directory in directories:
                if (
                    os.path.exists(os.path.join(directory, "pyproject.toml"))
                    and directory not in excluded_directories
                ):
                    subProjects.append(
                        Project(
                            project_name=os.path.basename(directory),
                            project_root=directory,
                            project_tool="uv.workspace",
                        )
                    )

    return subProjects


def find_sub_projects_in_project_dependencies(
    tomlData: dict[str, Any], toml_path: str
) -> list[Project]:
    """
    Finds sub projects in a project dependencies by looking for dependencies with
    the @ file:///${PROJECT_ROOT}//${SUBPROJECT_PATH} format:
        [project]
        dependencies = [
            "service-a @ file:///${PROJECT_ROOT}//services/service-a",
            "service-b @ file:///${PROJECT_ROOT}//services/service-b"
        ]
    """
    subProjects: list[Project] = []
    for dep in tomlData["project"]["dependencies"]:
        if isinstance(dep, str) and dep.find("@ file:///${PROJECT_ROOT}//"):
            rel_path = dep.split("@ file:///${PROJECT_ROOT}//")[1]
            subProjects.append(
                Project(
                    project_name=os.path.basename(dep),
                    project_root=get_absolute_path(toml_path, rel_path),
                    project_tool="project.dependencies",
                )
            )
    return subProjects


def get_absolute_path(path_to_root_toml: str, path_to_sub_project: str) -> str:
    path_to_root = os.path.dirname(path_to_root_toml)
    return os.path.abspath(os.path.join(path_to_root, path_to_sub_project))


def find_python_files(base_dir: str) -> List[str]:
    """
    Find all python files in the given directory and its subdirectories
    """
    python_files = []
    for root, dirs, files in os.walk(base_dir):
        for file in files:
            if file.endswith(".py"):
                python_files.append(os.path.join(root, file))
    return python_files


def read_file_contents(path: str) -> StringIO:
    """
    Read the contents of the file at the given path
    """
    with open(path, "r", newline="", encoding="utf-8") as file:
        return StringIO(file.read())


def parse_python_sources(paths: List[str]) -> List[SourceFile]:
    """
    Parse the given python files and return a list of SourceFile objects
    """
    parser = PythonParserBuilder().build()
    ctx = InMemoryExecutionContext()

    iterable_source_files: Iterable[SourceFile] = parser.parse_inputs(
        [
            ParserInput(
                Path(path),
                None,
                True,
                lambda: read_file_contents(path),
            )
            for path in paths
        ],
        None,
        ctx,
    )

    return list(iterable_source_files)



from io import BytesIO
import traceback
import socket

import cbor2

from rewrite_remote import RemotingContext, RemotingMessageType, OK, ERROR


def hello_world_handler(
    stream: BytesIO, sock: socket.socket, remoting_ctx: RemotingContext
) -> None:
    """
    A simple handler that responds to hello world command and response with string world
    """
    try:
        remoting_ctx.reset()
        request = cbor2.load(stream)

        if request != "hello":
            print("Did not receive 'hello' ")
            raise ValueError(f"Unexpected request: {request}")

        # Prepare a response
        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(OK, response_stream)
        cbor2.dump("world", response_stream)
        sock.sendall(response_stream.getvalue())

    except (socket.error, cbor2.CBORDecodeError) as e:
        print(f"Error in hello handler: {e}")
        traceback.print_exc()
        # Send an error response if something goes wrong
        response_stream = BytesIO()
        cbor2.dump(RemotingMessageType.Response, response_stream)
        cbor2.dump(ERROR, response_stream)
        cbor2.dump(traceback.format_exc(), response_stream)
        sock.sendall(response_stream.getvalue())



from typing import List, Optional, Dict, Any
from dataclasses import dataclass, field


@dataclass
class PackageSource:
    source: str = ""
    credential: Optional[Dict[str, str]] = None


@dataclass
class RecipeOption:
    name: str
    type: str
    required: bool
    value: Optional[Any]


@dataclass
class Recipe:
    name: str
    source: str
    options: List[RecipeOption] = field(default_factory=list)


@dataclass
class RecipeInstallResponse:
    recipes: List[Recipe]
    repository: str
    version: str



import os
import sys
import subprocess
import importlib
import inspect
import pkgutil
import site

from dataclasses import dataclass

from pypi_simple import PyPISimple
from packaging.version import parse as parse_version
from packaging.specifiers import SpecifierSet
from rewrite import Recipe

from typing import Optional, Dict, List, Any

DEFAULT_PYPI_URL = "https://pypi.org/simple"
DEFAULT_RECIPE_INSTALL_LOCATION = os.path.join(".", ".local_python_recipes")

# Ensure that the directory is on sys.path
if DEFAULT_RECIPE_INSTALL_LOCATION not in sys.path:
    sys.path.insert(0, DEFAULT_RECIPE_INSTALL_LOCATION)


class Source:
    def __init__(
        self,
        source: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ):
        self.source = source
        self.username = username
        self.password = password
        self.token = token


@dataclass
class Option:
    name: str
    type: str
    required: bool


@dataclass
class InstalledRecipe:
    name: str
    source: str
    options: List[Option]


class InstalledPackage:
    def __init__(
        self,
        name: str,
        version: str,
        source: str,
        recipes: List[InstalledRecipe],
    ):
        self.name = name
        self.version = version
        self.source = source
        self.recipes = recipes


class PyPiManager:
    @staticmethod
    def find_valid_source(
        package_name: str,
        requestedVersion: str,
        package_sources: List[Source],
        include_default_source: bool = True,
    ) -> Optional[Source]:
        """
        Finds a valid source for the specified package using `pip show` with `--index-url`.
        """
        package_identifier = (
            f"{package_name}=={requestedVersion}" if requestedVersion else package_name
        )
        if include_default_source and not any(
            source.source == DEFAULT_PYPI_URL for source in package_sources
        ):
            package_sources.append(Source(source=DEFAULT_PYPI_URL))

        for source in package_sources:
            try:
                authenticated_url = PyPiManager._get_authenticated_url(source)
                result = PyPiManager._package_exists_in_registry(
                    authenticated_url,
                    package_name,
                    requestedVersion,
                )
                if result:
                    print(f"Package {package_identifier} found in source: {source.source}")
                    return source
                else:
                    print(f"Package {package_identifier} not found in source: {source.source}")
            except Exception as e:
                print(f"Error checking source {source.source}: {e}")

        return None

    @staticmethod
    def install_package(
        package_name: str,
        requestedVersion: Optional[str] = None,
        package_source: Optional[Source] = None,
    ) -> InstalledPackage:
        """
        Installs the specified package from a given source.
        """
        package_identifier = (
            f"{package_name}=={requestedVersion}" if requestedVersion else package_name
        )

        pip_command = [
            sys.executable,
            "-m",
            "pip",
            "install",
            package_identifier,
            "--target",
            DEFAULT_RECIPE_INSTALL_LOCATION,
        ]

        if package_source:
            authenticated_url = PyPiManager._get_authenticated_url(package_source)
            pip_command.extend(["--index-url", authenticated_url])
        else:
            pip_command.extend(["--index-url", DEFAULT_PYPI_URL])

        # create directory if it does not exist
        if not os.path.exists(DEFAULT_RECIPE_INSTALL_LOCATION):
            os.makedirs(DEFAULT_RECIPE_INSTALL_LOCATION)

        try:
            subprocess.run(pip_command, check=True)
            metadata = PyPiManager._get_package_metadata(package_name)
            resolvedVersion = metadata.get("version", "")

            discovered_recipes = PyPiManager._introspect_module(package_name)

            return InstalledPackage(
                name=package_name,
                version=resolvedVersion,
                source=package_source.source if package_source else DEFAULT_PYPI_URL,
                recipes=discovered_recipes,
            )
        except Exception as e:
            raise RuntimeError(f"Failed to install package {package_name}: {e}")

    @staticmethod
    def uninstall_package(package_name: str) -> None:
        """
        Uninstalls the specified package using pip.
        """
        try:
            subprocess.run(["pip", "uninstall", "-y", package_name], check=True)
            print(f"Package {package_name} uninstalled successfully.")
        except Exception as e:
            raise RuntimeError(f"Failed to uninstall package {package_name}: {e}")

    @staticmethod
    def load_recipe(recipe_name: str, module_name: str, recipe_options: List[Option]) -> Recipe:
        """
        Loads a recipe from the specified source.
        """
        try:
            module = importlib.import_module(module_name)
            recipe = getattr(module, recipe_name)

            has_params = bool(inspect.signature(recipe).parameters)

            if not has_params:
                return recipe()
            else:
                return recipe(recipe_options)
        except Exception as e:
            raise RuntimeError(f"Failed to load recipe {recipe_name}: {e}")

    @staticmethod
    def load_package_details(package_name: str) -> Dict[str, Any]:
        """
        Loads package details using `pip show`.
        """
        try:
            metadata = PyPiManager._get_package_metadata(package_name)
            if not metadata:
                raise RuntimeError(f"Package {package_name} is not installed.")
            return metadata
        except Exception as e:
            raise RuntimeError(f"Failed to load package details for {package_name}: {e}")

    @staticmethod
    def _get_package_metadata(package_name: str) -> Dict[str, Any]:
        """
        Extracts metadata for an installed package using `pip show`.
        """
        # Ensure the custom install location is on sys.path
        if DEFAULT_RECIPE_INSTALL_LOCATION not in sys.path:
            sys.path.insert(0, DEFAULT_RECIPE_INSTALL_LOCATION)

        try:
            dist = importlib.metadata.distribution(package_name)
            # Convert it into a dictionary of metadata fields
            metadata_dict = {key.lower(): value for key, value in dist.metadata.items()}
            return metadata_dict
        except importlib.metadata.PackageNotFoundError:
            print(f"Package {package_name} not found in {DEFAULT_RECIPE_INSTALL_LOCATION}")
            return {}

    @staticmethod
    def _introspect_module(module_name: str) -> List[InstalledRecipe]:
        # Ensure the custom install location is on sys.path
        if DEFAULT_RECIPE_INSTALL_LOCATION not in sys.path:
            sys.path.insert(0, DEFAULT_RECIPE_INSTALL_LOCATION)

        # Convert to snake case per proper import conventions
        module_name = module_name.replace("-", "_")

        try:
            site.main()  # We want the recently installed module to be available to introspection
            module = importlib.import_module(module_name)
            submodules = [name for _, name, _ in pkgutil.iter_modules(module.__path__)]

            discovered_recipes: List[InstalledRecipe] = []

            for submodule in submodules:
                full_submodule_name = f"{module_name}.{submodule}"
                # Import each submodule
                sm = importlib.import_module(full_submodule_name)

                # Get all classes in the submodule
                classes = inspect.getmembers(sm, inspect.isclass)

                for class_name, class_obj in classes:
                    # Check if class is a subclass of Recipe
                    if issubclass(class_obj, Recipe) and class_obj is not Recipe:
                        discovered_recipes.append(
                            InstalledRecipe(
                                name=class_name,
                                source=full_submodule_name,
                                options=[],  # TODO support options
                            )
                        )

            return discovered_recipes
        except Exception as e:
            raise RuntimeError(f"Failed to introspect module {module_name}: {e}")

    @staticmethod
    def _package_exists_in_registry(
        index_url: str, package_name: str, version: Optional[str] = None
    ) -> bool:
        if not index_url.endswith("/"):
            index_url += "/"
        client = PyPISimple(endpoint=index_url)

        try:
            project_page = client.get_project_page(package_name)
            if not project_page:
                return False

            # When no version the presence of the package is enough to confirm a valid source
            if version is None:
                return True

            specifier = SpecifierSet(f"=={version}")

            # Hunt for a distribution that matches the requested version
            for dist in project_page.packages:
                if dist.version is not None:
                    dist_version = parse_version(dist.version)
                    if dist_version in specifier:
                        return True

            # No distributions matched the requested version specifier
            return False

        except Exception as e:  # pylint: disable=broad-except
            print(f"Error checking package {package_name}: {e}")
            return False

    @staticmethod
    def _get_authenticated_url(source: Source) -> str:
        """
        Returns the source URL with embedded authentication if username/password or token is provided.
        """
        if source.username and source.password:
            return source.source.replace(
                "https://", f"https://{source.username}:{source.password}@"
            )
        elif source.token:
            return source.source.replace("https://", f"https://{source.token}@")
        return source.source
